

#include "GetOptions.h"

GetOptions::GetOptions()
{
    Function* func = new Function;
    func->identifier = "DEFAULT";
    func->callName = " ";
    func->isSet = true;
    func->getopts = this;
    functionMap["DEFAULT"] = func;

#ifdef DEBUG
    dbg.info("Debug mode activated");
#endif

}

/*
//	addOption
//		adds a command line option to this command line option interpreter
//		-inputs - id - the identifier handle for this option
//				- sOp - the short form of the option
//				- lOp - the long form of the option
//				- tp - the type of value that this option uses
//		-return - 0 if it was successful
*/
int GetOptions::addOption(const char* id, const char* sOp, const char* lOp, type_id tp)
{

#ifdef DEBUG
    dbg << "Adding option: " << id << endl;
    dbg << "\tshort: " << sOp << endl;
    dbg << "\tlong: " << sOp << endl;
#endif

    Option* opt = new Option;
    opt->identifier = id;
    opt->shortOpt = sOp;
    opt->longOpt = lOp;
    opt->type = tp;
    if (tp == option::boolType)
        opt->value = "false";
    else
        opt->value = "";
    optionMap[string(id)] = opt;
    if (sOp != NULL)
    {
        optionToIdMap[string(sOp)] = string(id);
    }
    if (lOp != NULL)
    {
        optionToIdMap[string(lOp)] = string(id);
    }

#ifdef DEBUG
    dbg << "Finished adding option" << endl;
#endif

    return 0;
}

/*
//	addFunction
//		adds a command line function to this interpreter
//		-inputs - id - the identifier of the funciton
//				- callN - the call name of the function to be used in the command line to call the function
//		-reutrn - 0 if successful
//				- 1 if there is already a function with this id
//				- 2 if there is already a function with this call name
*/
int GetOptions::addFunction(const char* id, const char* callN)
{
    return addFunction(id, callN, new GetOptions());
}

/*
//	addFunction
//		adds a command line function to this interpreter
//		-inputs - id - the identifier of the funciton
//				- callN - the call name of the function to be used in the command line to call the function
//				- getopts - the GetOptions class to set to this function
//		-return - 0 if successful
//				- 1 if there is already a function with this id
//				- 2 if there is already a function with this call name
*/
int GetOptions::addFunction(const char* id, const char* callN, GetOptions* getopts)
{
    if (isFunc(id))
        return 1;
    if (isCallName(callN))
        return 2;
    Function* func = new Function;
    func->identifier = id;
    func->callName = callN;
    func->isSet = false;
    func->getopts = getopts;
    functionMap[string(id)] = func;
    callToFuncIdMap[string(callN)] = string(id);
    return 0;
}

/*
//	isFunctionActive
//		checks if the function identified by func is active, i.e. the user entered it in command line
//		-inputs - func - id of the function to check
//		-return - true if the function is active, false otherwise
*/
bool GetOptions::isFunctionActive(const char* func)
{
    if (!isFunc(func))
    {
        //ERROR STUFF
        return false;
    }
    return functionMap[string(func)]->isSet;
}


/*
//	setOptsToFunction
//		sets the GetOptions object associated with this function
//		-inputs - func - id of the function
//				- getopts - the GetOptions object which is set to this function
//		-return - 0 if successful
//				- 1 if the function doesn't exist
*/
int GetOptions::setOptsToFunction(const char* func, GetOptions* getopts)
{
    if (!isFunc(func))
    {
        //ERROR STUFF
        return 1;
    }
    delete functionMap[string(func)]->getopts;
    functionMap[string(func)]->getopts = getopts;
    return 0;
}

/*
//	getFuncOpts
//		get the GetOptions object of the function
//		-inputs - func - the identifier for the function to retrieve
//		-return - a pointer to the GetOptions object of the function,
///				Or NULL if the funciton doesn't exist
*/
GetOptions* GetOptions::getFuncOpts(const char* func)
{
    if (!isFunc(func))
        return NULL;
    return functionMap[string(func)]->getopts;
}


/*
//	parse
//		parses the string for any and all options, placing the values in the corresponding options in the map
//		-inputs - str - the string of options to be parsed
//		-return - 0 if it successfully identified and parsed the whole string for options
//				- 1 if the string contained something that couldn't be parsed
//				- 2 if the formatting of an option was wrong
//				- 3 if a value could not be set...most likely because of an invalid value
//				- 4 if there were problems with a function name or id
*/
int GetOptions::parse(const char* str)
{

#ifdef DEBUG
    dbg << "Parsing...." << endl;
#endif

    int index = 0;
    int end = strlen(str);
    int state = 0;
    int i;
    char* curOpt = new char[50];
    char* curOptVal = new char[200];
    while (index < end)
    {

#ifdef DEBUG
        dbg << "\tindex:" << index << ", state: " << state << endl;
#endif

        switch (state)
        {
        case 0:
            if (str[index] == ' ')
            {
                break;
            }
            if (str[index] == '-')
            {

#ifdef DEBUG
                dbg << "\tstate_change: 0 to 1" << endl;
#endif

                state = 1;
                break;
            }
            if ((str[index] >= 'a' && str[index] <= 'z') ||
                    (str[index] >= 'A' && str[index] <= 'Z'))
            {

#ifdef DEBUG
                dbg << "\tfunction..." << endl;
#endif

                i = 0;
                while (index < end && ((str[index] >= 'a' && str[index] <= 'z') ||
                                       (str[index] >= 'A' && str[index] <= 'Z') ||
                                       (str[index] >= '0' && str[index] <= '9') || str[index] == '-') && i < 49)
                {
                    curOpt[i] = str[index];
                    i++;
                    index++;
                }
                curOpt[i] = '\0';

#ifdef DEBUG
                dbg << "\tcurOpt: " << curOpt << endl;
#endif

                if (!isCallName(curOpt))
                    return 1;

#ifdef DEBUG
                dbg << "\tis call name" << endl;
#endif

                string funcId = callToFuncIdMap[string(curOpt)];
                if (!isFunc(funcId.c_str()))
                    return 4;

#ifdef DEBUG
                dbg << "\tit's a valid function" << endl;
#endif

                functionMap[funcId]->isSet = true;
                return functionMap[funcId]->getopts->parse(str + index);
            }
            return 1;
        case 1:
            curOpt[0] = '-';
            if (str[index] == '-')
            {

#ifdef DEBUG
                dbg << "\tstate_change: 1 to 2" << endl;
#endif

                state = 2;
                break;
            }
            i = 1;
            while (index < end && ((str[index] >= 'a' && str[index] <= 'z') ||
                                   (str[index] >= 'A' && str[index] <= 'Z') ||
                                   (str[index] >= '0' && str[index] <= '9') || str[index] == '-') && i < 49)
            {
                curOpt[i] = str[index];
                i++;
                index++;
            }
            curOpt[i] = '\0';

#ifdef DEBUG
            dbg << "\tcurOpt: " << curOpt << endl;
#endif

            if (!checkValidOp(curOpt))
                return 1;
            if (getType(curOpt) == option::boolType)
            {

#ifdef DEBUG
                dbg << "\tboolvalue" << endl;
#endif

                if (setValue(curOpt, "true") != 0)
                    return 3;
                else
                {

#ifdef DEBUG
                    dbg << "\tstate_change: 1 to 0" << endl;
#endif

                    state = 0;
                    break;
                }
            }

            if (str[index] != ' ')
            {
                return 2;
            }

#ifdef DEBUG
            dbg << "\tstate_change: 1 to 3" << endl;
#endif

            state = 3;
            break;
        case 2:
            curOpt[1] = '-';
            i = 2;
            while (index < end && ((str[index] >= 'a' && str[index] <= 'z') ||
                                   (str[index] >= 'A' && str[index] <= 'Z') || str[index] == '-') && i < 49)
            {
                curOpt[i] = str[index];
                i++;
                index++;
            }
            curOpt[i] = '\0';

#ifdef DEBUG
            dbg << "\tcurOpt: " << curOpt << endl;
#endif

            if (!checkValidOp(curOpt))
                return 1;
            if (getType(curOpt) == option::boolType)
            {

#ifdef DEBUG
                dbg << "\tboolvalue" << endl;
#endif

                if (setValue(curOpt, "true") != 0)
                    return 3;
                else
                {

#ifdef DEBUG
                    dbg << "\tstate_change: 3 to 0" << endl;
#endif

                    state = 0;
                    break;
                }
            }

            if (str[index] != '=')
            {
                return 2;
            }

#ifdef DEBUG
            dbg << "\tstate_change: 2 to 3" << endl;
#endif

            state = 3;
            break;
        case 3:
            i = 0;
            while (index < end && i < 199 &&
                    (str[index] != ' ' || (index < (end - 1) && str[index + 1] != '-')))
            {
                curOptVal[i] = str[index];
                i++;
                index++;
            }
            curOptVal[i] = '\0';

#ifdef DEBUG
            dbg << "\tcurOptVal: " << curOptVal << endl;
#endif

            if (setValue(curOpt, curOptVal) != 0)
                return 3;
            else
            {

#ifdef DEBUG
                dbg << "\tstate_change: 3 to 0" << endl;
#endif

                state = 0;
                break;
            }
        default:
            //ummmm
            break;
        }
        index++;
    }
    return 0;
}



/*
//	_type_Value
//		returns the value of the option "id" assuming that the function type corresponds to the option's value type
//		-inputs - id - the identifier of the option to get the value from
//		-return - the value of the option id, assuming the function type corresponds to the option value type
*/
int GetOptions::intValue(const char* id)
{
#ifdef DEBUG
    dbg << "Func_intValue: " << endl;
    dbg << "\tID: " << id << endl;
    dbg << "\tOption type: " << optionMap[string(id)]->type << endl;
#endif
    if (optionMap.count(string(id)) == 0)
    {

#ifdef DEBUG
        dbg << "ERROR: option not found" << endl;
#endif

        //ERROR STUFF
        return 0;
    }
    if (optionMap[string(id)]->type != option::intType)
    {

#ifdef DEBUG
        dbg << "ERROR: wrong option type" << endl;
#endif

        //ERROR STUFF
        return 0;
    }

#ifdef DEBUG
    dbg << "\tNo Errors thus far" << endl;
    dbg << "\tThe value for the option: " << optionMap[string(id)]->value.c_str() << endl;
    dbg << "\tOption type: " << optionMap[string(id)]->type << endl;
#endif

    istringstream strm(string(optionMap[string(id)]->value));
    int ret;
    strm >> ret;

#ifdef DEBUG
    dbg << "\tThe return value: " << ret << endl;
#endif

    return ret;
}
const char* GetOptions::cStrValue(const char* id)
{
    if (optionMap.count(string(id)) == 0)
    {
        //ERROR STUFF
        return NULL;
    }
    if (!(optionMap[string(id)]->type & (option::stringType | option::pathType)))
    {
        //ERROR STUFF
        return NULL;
    }
    return optionMap[string(id)]->value.c_str();
}
double GetOptions::doubleValue(const char* id)
{
    if (optionMap.count(string(id)) == 0)
    {
        //ERROR STUFF
        return 0;
    }
    if (optionMap[string(id)]->type != option::doubleType)
    {
        //ERROR STUFF
        return 0;
    }
    istringstream strm(string(optionMap[string(id)]->value));
    double ret;
    strm >> ret;
    return ret;
}

bool GetOptions::boolValue(const char* id)
{
    string ids(id);
    if (!isOp(id))
    {

#ifdef DEBUG
        dbg.error("not correct op id");
#endif
        //ERROR STUFF
        return false;
    }
    if (optionMap[ids]->type != option::boolType)
    {

#ifdef DEBUG
        dbg.error("not correct value type");
#endif
        //ERROR STUFF
        return false;
    }

#ifdef DEBUG
    dbg << "\t\t\tboolValue: " << optionMap[ids]->value.c_str() << endl;
#endif

    if (optionMap[ids]->value.compare("true") == 0)
    {
        return true;
    }
    else
    {
        return false;
    }
}

/*
//	testValidOptionValue
//		tests whether the value string contains a valid value for this value type
//		-inputs - val - the string containing the value to be tested
//				- type - the value type to be tested for
//		-return - true if the value is valid for this value type, false otherwise
*/
bool GetOptions::testValidOptionValue(string val, type_id type)
{
    istringstream strm(val);
    if (type == option::intType)
    {
        int test;
        strm >> test;
        if (strm.fail())
            return false;
        else
            return true;
    }
    if (type == option::doubleType)
    {
        double test;
        strm >> test;
        if (strm.fail())
            return false;
        else
            return true;
    }
    if (type == option::boolType)
    {
        if (val.compare("true") == 0 || val.compare("false"))
        {
            return true;
        }
        return false;
    }
    if (type == option::pathType)
    {
        return isValidPathName(val);
    }
    if (type == option::stringType)
    {
        //Not sure what to test right here....
    }
    return true;
}


/*
//	setValue
//		sets the value of an the option
//		-inputs - opt - the string taken from the command line which identifies the option
//				- val - the value to be set to the option
//		-return - an integer indicating success or not, 0 means successful
*/
int GetOptions::setValue(const char* opt, const char* val)
{

#ifdef DEBUG
    dbg << "\t\tOption: " << opt << ", Value: " << val << endl;
#endif

    if (!checkValidOp(opt))
        return 1;
    string opId = optionToIdMap[string(opt)];
    if (!testValidOptionValue(string(val), getType(opt)))
        return 2;
    optionMap[opId]->value.assign(val);
    return 0;
}

/*
//	checkValidOp
//		checks whether the string refers to a valid option
//		-inputs - opt - the string, which will be tested
//		-return - true if it's a valid option, false otherwise
*/
bool GetOptions::checkValidOp(const char* opt)
{
    if (optionToIdMap.count(opt) > 0)
    {
        return true;
    }
    else
    {
        return false;
    }
}

/*
//	checkType
//		retrieves the value type specified for an option
//		-inputs - opt - the option string, that identifies the option
//		-return - the value type of the option or 0 if it isn't a valid option string
*/
type_id GetOptions::getType(const char* opt)
{
    if (!checkValidOp(opt))
    {
        return 0;
    }
    else
    {
        return optionMap[optionToIdMap[string(opt)]]->type;
    }
}


/*
//	isValidPathName
//		tests if the string is a valid path in this operating system
//		-inputs - path - the string to be validated as a path
//		-return - true if the string is a valid path, false otherwise
*/
bool GetOptions::isValidPathName(string path)
{
#ifdef WIN32
    //stuff for windows
    return true;
#elif defined UNIX
    //stuff for unix
    return true;
#elif defined MAC
    //stuff for macs
    return true;
#else
    return true;
#endif
}


/*
//	isFunc
//		checks if func is an id for a function that has been added to
//		-inputs - func - identifier of the function
//		-return - true if the function exists, false otherwise
*/
bool GetOptions::isFunc(const char* func)
{
    if (functionMap.count(string(func)) > 0)
    {
        return true;
    }
    else
    {
        return false;
    }
}

/*
//	isOp
//		checks if id is an identifier for an option that already exists
//		-inputs - id - identifier of the option
//		-return - true if the option exists, false otherwise
*/
bool GetOptions::isOp(const char* id)
{
    if (optionMap.count(string(id)) > 0)
    {
        return true;
    }
    else
    {
        return false;
    }
}

/*
//	isCallName
//		checks whether the string refers to a call name for a function that exists
//		-inputs - opt - the string, which will be tested
//		-return - true if it's a call name, false otherwise
*/
bool GetOptions::isCallName(const char* callN)
{
    if (callToFuncIdMap.count(string(callN)) > 0)
    {
        return true;
    }
    else
    {
        return false;
    }
}

/*
//	isOpActive
//		checks if the option identified by op is active, i.e. the user entered it in command line
//		-inputs - op - id of the option to check
//		-return - true if the option is active, false otherwise
*/
bool GetOptions::isOpActive(const char* op)
{
    if (!isOp(op))
    {
        //ERROR STUFF
        return false;
    }
    else if (optionMap[string(op)]->value.empty())
    {
        return false;
    }
    return true;
}