#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "dtic.h"

#define MAX(x,y) ((x)>(y)?(x):(y))

#define alloc_error_check(p) {if((p)==NULL){fprintf(stderr,"Allocation Failure!\n");exit(1);}}

int writeFile_double(const char* filename, double *vec, int nDims, int *dims)
{
    int size = 1;
    int ii;
    FILE * theFile;

    printf(" -- Writing to %s \n", filename);
    if ((theFile = fopen(filename, "wb")) == NULL)
    {
        printf("ERROR: file: %s could not be opened.\n", filename);
        return 1;
    }

    for (ii = 0; ii < nDims; ii++)
    {
        size *= dims[ii];
    }

    fwrite(vec, sizeof(double), size, theFile);

    fclose(theFile);

    return 0;

}

int writeFile_double2text(const char *filename, double *vec, int nDims, int *dims, int flag)
{
    int rest = 1;
    int ii, jj;
    FILE *theFile;

    if ((flag != 0) && (flag != 1)) flag = 0;

    if (flag == 0)
    {
        printf("-- Writing to %s in text. Overwrite Mode.\n", filename);
        if ((theFile = fopen(filename, "w")) == NULL)
        {
            printf("ERROR: file: %s could not be opened.\n", filename);
            return 1;
        }
    }
    else if (flag == 1)
    {
        printf("-- Writing to %s in text. Append Mode.\n", filename);
        if ((theFile = fopen(filename, "a")) == NULL)
        {
            printf("ERROR: file: %s could not be opened.\n", filename);
            return 1;
        }

        fprintf(theFile, "\n ------- start appending ------ \n");
    }

    if (nDims >= 2)
    {
        for (ii = 1; ii < nDims; ++ii)
        {
            rest *= dims[ii];
        }
    }

    for (jj = 0;jj < rest;++jj)
    {
        for (ii = 0;ii < dims[0];++ii)
        {
            fprintf(theFile, "%12.4e,  ", vec[ii+ jj*dims[0]]);
        }
        fprintf(theFile, "\n");
    }

    fclose(theFile);

    return 0;

}

int writeFile_log(const char *filename, char *s, int flag)
{
    FILE *theFile;

    if ((flag != 0) && (flag != 1)) flag = 0;

    if (flag == 0)
    {
        printf("-- Writing to %s in text. Overwrite Mode.\n", filename);
        if ((theFile = fopen(filename, "w")) == NULL)
        {
            printf("ERROR: file: %s could not be opened.\n", filename);
            return 1;
        }
    }
    else if (flag == 1)
    {
        printf("-- Writing to %s in text. Append Mode.\n", filename);
        if ((theFile = fopen(filename, "a")) == NULL)
        {
            printf("ERROR: file: %s could not be opened.\n", filename);
            return 1;
        }

        fprintf(theFile, "\n ------- start appending ------ \n");
    }

    fprintf(theFile, "%s \n", s);

    fclose(theFile);

    return 0;

}

int readDifSchemFile(const char* filename, float difscheme[])
{
    FILE *fp = NULL;
    char currline[256];
    int  max_chars_per_line = 250;
    int  jj = 0;

    if ((fp = fopen(filename, "r")) == NULL)
    {
        return -1;
    }
    else
    {
        while (!feof(fp))
        {
            fgets(currline, max_chars_per_line, fp);
            sscanf(currline, "%f %f %f", &difscheme[3*jj], &difscheme[3*jj+1], &difscheme[3*jj+2]);
            jj++;
        }
        fclose(fp);
    }
    return jj;
}

void PhaseUnWrap1D(double Ph[], int len)
{

    double dp[1024], dp_delta[1024];
    int ii;

    for (ii = 0;ii < len - 1;++ii)
    {
        dp[ii] = Ph[ii+1] - Ph[ii];
        dp_delta[ii] = atan(sin(dp[ii]) / cos(dp[ii]));
    }

    for (ii = 1; ii < len;ii++)
    {
        Ph[ii] = Ph[ii-1] + dp_delta[ii-1];
    }
}

void LinearFit(double x[], double y[], int n, double a[], double dt[])
{
    int i;
    double xx, yy, e, f, q, u, p, umax, umin, s;

    xx = 0.0;
    yy = 0.0;
    for (i = 0; i <= n - 1; i++)
    {
        xx = xx + x[i] / n;
        yy = yy + y[i] / n;
    }
    e = 0.0;
    f = 0.0;
    for (i = 0; i <= n - 1; i++)
    {
        q = x[i] - xx;
        e = e + q * q;
        f = f + q * (y[i] - yy);
    }
    a[1] = f / e;
    a[0] = yy - a[1] * xx;
    q = 0.0;
    u = 0.0;
    p = 0.0;
    umax = 0.0;
    umin = 1.0e+30;
    for (i = 0; i <= n - 1; i++)
    {
        s = a[1] * x[i] + a[0];
        q = q + (y[i] - s) * (y[i] - s);
        p = p + (s - yy) * (s - yy);
        e = fabs(y[i] - s);
        if (e > umax) umax = e;
        if (e < umin) umin = e;
        u = u + e / n;
    }
    dt[1] = sqrt(q / n);
    dt[0] = q;
    dt[2] = p;
    dt[3] = umax;
    dt[4] = umin;
    dt[5] = u;
    return;
}

double dB1P(double te[], double ph1d[], int N)
{
    double delta;
    double coef[2], info[6];

    PhaseUnWrap1D(ph1d, N);
    LinearFit(te, ph1d, N, coef, info);
    delta = coef[1];
    return delta;
}

double *dB0Map(double *PhImVol, double TE[], int dx, int dy, int dz, int dt)
{
    int ii, jj, kk, tt;
    int ndx;
    int N;
    double bar[1024], barEven[512], barOdd[512], teEven[512], teOdd[512];
    double delta[2];
    double *dB0;
    double gyro = 2 * 42580000 * 3.14159265;

    if (dt < 8)
    {
        printf("\n Warning: Number of phase images at different TEs is less than 8.\n\n");
    }

    if ((dB0 = malloc(dx * dy * dz * sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        return dB0;
    }

    N = dt / 2;

    for (kk = 0; kk < dz; kk++)
    {
        for (jj = 0; jj < dy; jj++)
        {
            for (ii = 0; ii < dx; ii++)
            {
                for (tt = 0; tt < dt; tt++)
                {
                    ndx = tt * (dz * dy * dx) + kk * (dy * dx) + jj * (dx) + ii;
                    bar[tt] = PhImVol[ndx];
                }

                for (tt = 0; tt < N; tt++)
                {

                    barOdd[tt] = bar[tt*2];
                    barEven[tt] = bar[tt*2+1];
                    teOdd[tt] = TE[tt*2];
                    teEven[tt] = TE[tt*2+1];
                }

                delta[0] = dB1P(teOdd, barOdd, N);
                delta[1] = dB1P(teEven, barEven, N);

                ndx = kk * (dy * dx) + jj * (dx) + ii;
                dB0[ndx] = (delta[0] + delta[1]) / 2.0 / gyro;

            }

        }
    }
    return dB0;
}

void dB0Map_NoMemAlloc(double *dB0, double *PhImVol, double TE[], int dx, int dy, int dz, int dt)
{
    int ii, jj, kk, tt;
    int ndx;
    int N;
    double bar[1024], barEven[512], barOdd[512], teEven[512], teOdd[512];
    double delta[2];

    double gyro = 2 * 42580000 * 3.14159265;

    if (dt < 8)
    {
        printf("\n Warning: Number of phase images at different TEs is less than 8.\n\n");
    }

    N = dt / 2;

    for (kk = 0; kk < dz; kk++)
    {
        for (jj = 0; jj < dy; jj++)
        {
            for (ii = 0; ii < dx; ii++)
            {
                for (tt = 0; tt < dt; tt++)
                {
                    ndx = tt * (dz * dy * dx) + kk * (dy * dx) + jj * (dx) + ii;
                    bar[tt] = PhImVol[ndx];
                }

                for (tt = 0; tt < N; tt++)
                {

                    barOdd[tt] = bar[tt*2];
                    barEven[tt] = bar[tt*2+1];
                    teOdd[tt] = TE[tt*2];
                    teEven[tt] = TE[tt*2+1];
                }

                delta[0] = dB1P(teOdd, barOdd, N);
                delta[1] = dB1P(teEven, barEven, N);

                ndx = kk * (dy * dx) + jj * (dx) + ii;
                dB0[ndx] = (delta[0] + delta[1]) / 2.0 / gyro;

            }

        }
    }

}

double lgr1d(double x[], double y[], int n, double t)
{
    int i, j, k, m;
    double z, s;
    z = 0.0;
    if (n < 1) return(z);
    if (n == 1)
    {
        z = y[0];
        return(z);
    }
    if (n == 2)
    {
        z = (y[0] * (t - x[1]) - y[1] * (t - x[0])) / (x[0] - x[1]);
        return(z);
    }
    i = 0;
    while ((x[i] < t) && (i < n)) i = i + 1;
    k = i - 4;
    if (k < 0) k = 0;
    m = i + 3;
    if (m > n - 1) m = n - 1;
    for (i = k;i <= m;i++)
    {
        s = 1.0;
        for (j = k;j <= m;j++)
            if (j != i) s = s * (t - x[j]) / (x[i] - x[j]);
        z = z + s * y[i];
    }
    return(z);
}

void epiCorr(double ioEpi[], double b0Map[], double Tacq, int dx, int dy, int dz)
{
    int ii, kk, rr, nn;
    int ndx1, ndx2;
    double pX[1024], pY[1024], pXI[1024];
    const double gyrobar = 42580000;

    nn = 0;

    for (rr = 0;rr < dy;rr++)
    {
        pX[rr] = rr;
    }

    for (kk = 0;kk < dz;kk++)
    {
        printf("  kk= %d \n", kk);
        for (ii = 1;ii < dx - 1;ii++)
        {
            ndx1 = kk * (dx * dy) + ii;
            for (nn = 0;nn < dy;nn++)
            {
                ndx2 = ndx1 + nn * dx;
                pY[nn] = ioEpi[ndx2];

                pXI[nn] = pX[nn] + (gyrobar * Tacq * b0Map[ndx2]);
            }

            for (rr = 1;rr < dy - 1;rr++)
            {
                ndx2 = ndx1 + rr * dx;
                ioEpi[ndx2] = lgr1d(pX, pY, dy, pXI[rr]);
            }
        }
    }
}

double otsu(double *im, int rows, int cols, int x0, int y0, int dx, int dy, int vvv)
{
    unsigned char *np;
    int thresholdValue = 1;
    int ihist[256];
    int i, j, k;
    int n, n1, n2, gmin, gmax;
    double m1, m2, sum, csum, fmax, sb;

    unsigned char *image;
    double immax = 0;

    for (i = 0; i < rows * cols; ++i)
    {
        if (immax < im[i])
        {
            immax = im[i];
        }
    }

    if ((image = malloc(rows * cols * sizeof(unsigned char))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        return -1;
    }

    for (i = 0; i < rows * cols; ++i)
        image[i] = (unsigned char)(im[i] / immax * 255);

    memset(ihist, 0, sizeof(ihist));
    gmin = 255;
    gmax = 0;

    for (i = y0 + 1; i < y0 + dy - 1; i++)
    {
        np = &image[i*cols+x0+1];
        for (j = x0 + 1; j < x0 + dx - 1; j++)
        {
            ihist[*np]++;
            if (*np > gmax) gmax = *np;
            if (*np < gmin) gmin = *np;
            np++;
        }
    }

    sum = csum = 0.0;
    n = 0;
    for (k = 0; k <= 255; k++)
    {
        sum += (double) k * (double) ihist[k];
        n   += ihist[k];
    }
    if (!n)
    {

        fprintf(stderr, "NOT NORMAL thresholdValue = 160\n");
        free(image);

        return (immax*160.0 / 255.0);
    }

    fmax = -1.0;
    n1 = 0;
    for (k = 0; k < 255; k++)
    {
        n1 += ihist[k];
        if (!n1)
        {
            continue;
        }
        n2 = n - n1;
        if (n2 == 0)
        {
            break;
        }
        csum += (double) k * ihist[k];
        m1 = csum / n1;
        m2 = (sum - csum) / n2;
        sb = (double) n1 * (double) n2 * (m1 - m2) * (m1 - m2);

        if (sb > fmax)
        {
            fmax = sb;
            thresholdValue = k;
        }
    }

    if (vvv & 1)
    {
        fprintf(stderr, "# OTSU: thresholdValue = %d. gmin=%d, gmax=%d.\n",
                thresholdValue, gmin, gmax);
        printf("\n immax = %10.3f,    thres = %10.3f. \n", immax, immax*thresholdValue / 255.0);
    }
    free(image);

    return(immax*thresholdValue / 255.0);
}

int findFirstGT(double *vec, int n, double t)
{
    int ii = 0;

    while (ii < n)
    {
        if (vec[ii] < t)
        {
            ++ii;
        }
        else
        {
            break;
        }
    }

    return ii;
}

int findLastGT(double *vec, int n, double t)
{
    int ii;

    ii = n - 1;

    while (ii >= 0)
    {
        if (vec[ii] < t)
        {
            --ii;
        }
        else
        {
            break;
        }
    }

    return ii;
}

double *eddygrad(double *field, int dx, int dy, int dz)
{
    const double eps = 0.000000000000001;

    int i, vsize;
    double *grad;
    double *mask;

    vsize = dx * dy * dz;
    if ((mask = malloc(vsize * sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        return mask;
    }

    for (i = 0; i < vsize; ++i)
    {
        if ((fabs(field[i])) <= eps)
        {
            mask[i] = 0;
        }
        else
        {
            mask[i] = 1;
        }
    }

    grad = eddygrad_mask(field, dx, dy, dz, mask);

    free(mask);

    return grad;
}

double *eddygrad_mask(double *field, int dx, int dy, int dz, double * mask)
{
    double th = 0.000001;

    int i, k, ndx, ndx1, ndx2;
    int pos1, pos2, n;
    double *grad;

    double a[2], errbound[6];

    double *pnt;
    double f1[1024], f2[1024], f3[1024];

    if ((grad = malloc(4 * dz * sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        return grad;
    }

    for (k = 0; k < dz; ++k)
    {
        ndx = k * (dy * dx);

        ndx1 = (dy / 2) * dx + ndx;
        pnt = mask + ndx1;

        pos1 = findFirstGT(pnt, dx, th);
        pos2 = findLastGT(pnt, dx, th);

        if (pos2 <= pos1)
        {
            printf("Error: Edge detection on 1D failed. \n");
            free(grad);
            return NULL;
        }

        n = pos2 - pos1 + 1;

        for (i = 0; i < n;++i)
        {
            f1[i] = (pos1 + i) - (dx / 2) ;
            f2[i] = field[ndx1+pos1+i];

        }

        LinearFit(f1, f2, n, a, errbound);

        grad[4*k] = a[1];
        grad[4*k+1] = a[0];

        ndx2 = ndx + dx / 2;
        for (i = 0; i < dy;++i)
        {
            f1[i] = field[ndx2+i*dx];
            f3[i] = mask[ndx2+i*dx];
        }

        pos1 = findFirstGT(f3, dy, th);
        pos2 = findLastGT(f3, dy, th);
        if (pos2 <= pos1)
        {
            printf("Error: Edge detection on 1D failed. \n");
            free(grad);
            return NULL;
        }

        n = pos2 - pos1 + 1;

        for (i = 0; i < n;++i)
        {
            f2[i] = f1[pos1+i];
        }
        for (i = 0;i < n;++i)
        {
            f1[i] = pos1 + i - (dy / 2);
        }

        LinearFit(f1, f2, n, a, errbound);

        grad[4*k+2] = a[1];
        grad[4*k+3] = a[0];
    }

    return grad;
}

void reconGradField(double *vol, double *grad, int dx, int dy, int dz)
{
    int i, j, k, ndx;
    double a1, b1, a2, b2;
    int px, py;
    int midx, midy;

    midx = dx / 2;
    midy = dy / 2;

    for (k = 0; k < dz; ++k)
    {
        a1 = grad[4*k];
        b1 = grad[4*k +1];
        a2 = grad[4*k +2];
        b2 = grad[4*k +3];

        ndx = k * (dx * dy);
        for (j = 0; j < dy; ++j)
        {
            py = j - midy;
            for (i = 0; i < dx; ++i)
            {
                px = i - midx;

                vol[ndx+j*dx+i] = a1 * px + a2 * py + (b1 + b2) / 2.0;
            }
        }
    }
}

void reconEddyField(double* ecif, double* CalibData, double difdir[], int dx, int dy, int dz)
{}

void dtiCorrFull(double *ioDti, double *b0Map, double *eddyMap, double Tacq, int dx, int dy, int dz, int dw)
{
    int i, w;
    int ndx;

    for (i = 0; i < dx*dy*dz; ++i)
    {
        eddyMap[i] += b0Map[i];
    }

    for (w = 0; w < dw; ++w)
    {
        ndx = w * (dx * dy * dz);
        epiCorr(&(ioDti[ndx]), eddyMap, Tacq, dx, dy, dz);
    }
}

void dtiCorrB0(double *ioDti, double *b0Map, double Tacq, int dx, int dy, int dz, int dw)
{
    int w;
    int ndx;

    for (w = 0; w < dw; ++w)
    {
        ndx = w * (dx * dy * dz);
        epiCorr(&(ioDti[ndx]), b0Map, Tacq, dx, dy, dz);
    }
}

double *magMask2(double *Mag, int dx, int dy, int dz, int dt)
{
    int i, k, ndx;
    int mid = dt / 2;

    double *mask, *pnt;
    double lev;

    if ((mask = malloc(dx * dy * dz * sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        return mask;
    }

    ndx = (dt / 2) * (dx * dy * dz);
    for (i = 0;i < dx*dy*dz;++i)
    {
        mask[i] = Mag[ndx+i];
    }

    for (k = 0; k < dz; ++k)
    {
        ndx = k * (dy * dx);
        pnt = (mask + ndx);
        lev = otsu(pnt, dx, dy, 0, 0, dx, dy, 0);
        for (i = 0;i < dx*dy;++i)
        {
            if (pnt[i] < lev)
            {
                pnt[i] = 0;
            }
            else
            {
                pnt[i] = 1;
            }
        }
    }

    return mask;

}

void displayMatrix(double *inMat, int offset, int n)
{
    int ii;
    for (ii = offset; ii < offset + n;ii++)
    {
        printf("%12.4e,  ", inMat[ii]);
    }
    printf("\n");
}

double* dti_correction(double *Dti, double *dB_base, double *Calib, struct dtiP P)
{
    int i, r;
    int volsize;
    double *pnt, *pnt_grad;
    double  *dB_dwi;

    int dim[5];
    const char outfname1[] = ".\\dti_corrected.dat";
    const char outfname2[] = ".\\dti_orig.dat";
    const char outfname3[] = ".\\dB_base_in_dti_correction.dat";
    const char outfname4[] = ".\\log.txt";
    char strInfo[4096];

    dim[0] = P.dx;
    dim[1] = P.dy;
    dim[2] = P.dz;
    dim[3] = P.dw;
    writeFile_double(outfname2 , Dti, 4 , dim);
    writeFile_double(outfname3 , dB_base, 3 , dim);

    if (P.Tacq <= 0.001)
    {
        printf("Warning: Acq time is toooo short.\n");
        P.Tacq = 44.864;
    }

    P.Tacq = P.Tacq / 1000;

    sprintf(strInfo, "P.Tacq=%12.4e = %12.4f before 1/1000 \n", P.Tacq, P.Tacq);
    writeFile_log(outfname4, strInfo, 1);

    epiCorr(Dti, dB_base, P.Tacq, P.dx, P.dy, P.dz);

    volsize = P.dx * P.dy * P.dz;

    if ((dB_dwi = malloc(volsize*sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        return NULL;
    }

    if (Calib == NULL)
    {
        for (i = 1;i < P.dw;++i)
        {
            pnt = Dti + i * volsize;
            epiCorr(pnt, dB_base, P.Tacq, P.dx, P.dy, P.dz);
        }
    }
    else
    {
        for (i = 1; i  < P.dw; ++i)
        {

            pnt_grad = Calib + (i - 1) * 4 * P.dz;
            reconGradField(dB_dwi, pnt_grad, P.dx, P.dy, P.dz);
            for (r = 0;r < volsize; ++r)
            {
                dB_dwi[r] += dB_base[r];
            }

            pnt = Dti + i * volsize;
            epiCorr(pnt, dB_dwi, P.Tacq, P.dx, P.dy, P.dz);
        }
    }

    dim[0] = P.dx;
    dim[1] = P.dy;
    dim[2] = P.dz;
    dim[3] = P.dw;
    writeFile_double(outfname1 , Dti, 4 , dim);

    free(dB_dwi);
    return Dti;
}

double* dti_correction_3d(double *Dti, double *dB_base, double* CalibData, float dtiScheme[], struct dtiP P)
{
    int i, j, r;
    int volsize;
    double *pnt, *pnt_grad;
    double  *dB_dwi;
    int DW;
    float* difdir;
    int dims[4];

    if (P.Tacq <= 0.001)
    {
        printf("Warning: Acq time is toooo short.\n");
        P.Tacq = 44.864;
    }
    P.Tacq = P.Tacq / 1000;

    epiCorr(Dti, dB_base, P.Tacq, P.dx, P.dy, P.dz);

    volsize = P.dx * P.dy * P.dz;

    if ((dB_dwi = malloc(volsize*sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        return NULL;
    }

    DW = P.dw - 1;

    for (i = 1; i  < P.dw; ++i)
    {
        difdir = dtiScheme + 3 * (i - 1);
        for (j = 0;j < volsize;++j)
        {
            dB_dwi[j] = difdir[0] * CalibData[j] + difdir[1] * CalibData[j+volsize] + difdir[2] * CalibData[j+2*volsize] + CalibData[j+3*volsize];
        }

        for (r = 0;r < volsize; ++r)
        {
            dB_dwi[r] += dB_base[r];
        }

        pnt = Dti + i * volsize;
        epiCorr(pnt, dB_dwi, P.Tacq, P.dx, P.dy, P.dz);
    }

    free(dB_dwi);
    return Dti;

}

double* dti_calibration2(double *Phase, struct dtiP P, double *Mask)
{
    int i, k;
    int offset, ndx;
    double *Grad, *GradAll;
    double *dB_base, *dB_dwi;

    int dim[5] = {64, 64, 1, 64, 7};
    const char outfname[] = ".\\dB_base_from_calib2Func.dat";
    const char outfname1[] = ".\\dB_dwi_from_calib2Func.dat";

    if (P.dTE > 0)
    {
        for (i = 0; i < P.dt; ++i)
        {
            P.TE[i] = P.dTE * i;
        }
    }

    if ((Grad = malloc(4*P.dz*sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        return NULL;
    }

    if ((GradAll = malloc(4*P.dz*(P.dw - 1)*sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        free(Grad);
        return NULL;
    }

    dB_base = Phase;
    dB0Map_NoMemAlloc(dB_base, Phase, P.TE, P.dx, P.dy, P.dz, P.dt);

    writeFile_double(outfname, dB_base, 3, dim);

    dB_dwi = Phase + (P.dx*P.dy*P.dz);

    for (k = 1; k < P.dw; ++k)
    {
        ndx = k * (P.dx * P.dy * P.dz * P.dt);

        dB0Map_NoMemAlloc(dB_dwi, Phase + ndx, P.TE, P.dx, P.dy, P.dz, P.dt);

        if (k == 1) writeFile_double(outfname1, dB_dwi, 3, dim);

        for (i = 0;i < (P.dx*P.dy*P.dz);++i)
        {
            dB_dwi[i] = dB_dwi[i] - dB_base[i];
        }

        Grad = eddygrad_mask(dB_dwi, P.dx, P.dy, P.dz, Mask);

        offset = (k - 1) * (4 * P.dz);
        for (i = 0; i < (4* P.dz); ++i)
        {
            GradAll[offset + i] = Grad[i];
        }

    }

    free(Grad);
    return GradAll;

}

double* dti_calibration(double *FieldMap, struct dtiP P)
{
    int i, k;
    int offset;
    double *Grad, *GradAll;
    double *dB_base, *dB_dwi;

    int dim[8];
    const char outfname1[] = ".\\dB_AllDir_rewrite.dat";
    const char outfname3[] = ".\\grad_dti_calib.txt";
    const char logfilename[] = ".\\log.txt";
    char strInfo[4096];

    writeFile_log(logfilename, "Enter dti_calibration.", 1);
    dim[0] = P.dx;
    dim[1] = P.dy;
    dim[2] = P.dz;
    dim[3] = P.dw;
    sprintf(strInfo, "dim=[%d,%d,%d,%d], dTE=%12.4e \n", dim[0], dim[1], dim[2], dim[3], P.dTE);
    writeFile_log(logfilename, strInfo, 1);
    writeFile_double(outfname1, FieldMap, 4, dim);

    writeFile_log(logfilename, "Before alloc Grad.", 1);

    if ((Grad = malloc(4*P.dz*sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        return NULL;
    }

    if ((GradAll = malloc(4*P.dz*(P.dw - 1)*sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        free(Grad);
        return NULL;
    }

    dB_base = FieldMap;

    writeFile_log(logfilename, "Before calc Grad.", 1);

    for (k = 1; k < P.dw; ++k)
    {
        dB_dwi = FieldMap + k * (P.dx * P.dy * P.dz);

        for (i = 0;i < (P.dx*P.dy*P.dz);++i)
        {
            dB_dwi[i] = dB_dwi[i] - dB_base[i];
        }

        Grad = eddygrad(dB_dwi, P.dx, P.dy, P.dz);

        offset = (k - 1) * (4 * P.dz);
        for (i = 0; i < (4* P.dz); ++i)
        {
            GradAll[offset + i] = Grad[i];
        }

    }

    dim[0] = 4;
    dim[1] = P.dz *(P.dw - 1);
    writeFile_double2text(outfname3, GradAll, 2, dim, 0);

    free(Grad);
    return GradAll;

}

void CDFFT2_fftshift(double *im, int dx, int dy)
{
    int *ip, n1, n2, n, i, j;
    double **a, *w, err;
    double *temp;

    n1 = dy;
    n2 = dx * 2;

    n = MAX(n1, n2 / 2);
    ip = alloc_1d_int(2 + (int) sqrt(n + 0.5));
    n = MAX(n1, n2) * 3 / 2;
    w = alloc_1d_double(n);
    ip[0] = 0;

    a = (double **) malloc(sizeof(double *) * n1);
    alloc_error_check(a);
    a[0] = im;
    for (j = 1; j < n1; j++)
    {
        a[j] = a[j - 1] + n2;
    }

    cdft2d(n1, n2, -1, a, NULL, ip, w);

    temp = malloc(sizeof(double) * dx);
    alloc_error_check(temp);

    for (i = 0;i < dy;++i)
    {
        memcpy(temp, (im + n2*i), sizeof(double)*dx);
        memcpy((im + n2 * i), (im + n2*i + dx), sizeof(double)*dx);
        memcpy((im + n2*i + dx), temp, sizeof(double)*dx);
    }

    free_1d_double(w);
    free_1d_int(ip);
    free(a);
    free(temp);
}

void epcsi2d_conj(short *slice, struct dtiP P, double *ImComplex3d)
{
    int onerow;
    long ndx1 = 0, ndx2 = 0;
    double *ImComplex2d;
    int i, j, t;

    onerow = 2*((P.dx + TRANS_POINTS)*P.dt - TRANS_POINTS);

    for (j = 0; j < P.dy; ++j)
    {
        for (t = 0; t < P.dt; ++t)
        {
            ndx2 = j * onerow + t * 2 * (P.dx + TRANS_POINTS);
            ndx1 = t * 2 * P.dx * P.dy + j * 2 * P.dx;

            if ((t % 2) == 0)
            {
                for (i = 0; i < 2*P.dx; ++i)
                {
                    ImComplex3d[ndx1+2*i] = (double) slice[ndx2+2*i];
                    ImComplex3d[ndx1+2*i+1] = -((double) slice[ndx2+2*i+1]);
                }
            }
            else
            {

                for (i = 0; i < P.dx; ++i)
                {
                    ImComplex3d[ndx1+2*i] = (double) slice[ndx2+2*P.dx - 2*(i+1)];
                    ImComplex3d[ndx1+2*i+1] = -((double) slice[ndx2+2*P.dx - 2*(i+1)+1]);
                }
            }
        }
    }

    for (i = 0; i < P.dt; ++i)
    {

        ImComplex2d = ImComplex3d + i * (2 * P.dx * P.dy);
        CDFFT2_fftshift(ImComplex2d, P.dx, P.dy);
    }

}

void B2LInt16(short *s, int n)
{
    int i;
    unsigned char c1, c2;
    unsigned char *p;

    for (i = 0; i < n; ++i)
    {
        p = (char *)s;
        c1 = p[2*i] & 0xFF;
        c2 = p[2*i+1] & 0xFF;
        s[i] = ((short) c1 << 8) + c2;
    }
}

short * readGEPfile_int16B(const char* Pfilename, struct dtiP P)
{
    int onerow, daq_junk_bytes, size;

    short *vec;
    FILE * theFile;

    if ((theFile = fopen(Pfilename, "rb")) == NULL)
    {
        printf("ERROR: file: %s could not be opened.\n", Pfilename);
        return NULL;
    }

    onerow = 2*((P.dx + TRANS_POINTS)*P.dt - TRANS_POINTS);
    size = onerow *P.dy *P.dz*P.dw;

    daq_junk_bytes = onerow *2;

    fseek(theFile, GE_HDR_SIZE + daq_junk_bytes, SEEK_SET);

    if ((vec = malloc(size*sizeof(short))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        fclose(theFile);
        return vec;
    }

    fread(vec, sizeof(short), size, theFile);
    B2LInt16(vec, size);
    fclose(theFile);

    return vec;
}

void phaserecon(short* vec, struct dtiP P, double * phase5d, double* mag5d)
{

    int onerow, oneslice;
    int z, w, t, y, x;
    int ndx1, ndx2;
    short *ndxSlice;
    double *cmplxSlice;

    onerow = 2*((P.dx + TRANS_POINTS)*P.dt - TRANS_POINTS);
    oneslice = onerow* P.dy;

    cmplxSlice = malloc(2*P.dx*P.dy*P.dt * sizeof(double));
    alloc_error_check(cmplxSlice);

    for (z = 0;z < P.dz; ++z)
    {
        for (w = 0;w < P.dw; ++w)
        {
            ndxSlice = vec + oneslice * w + (oneslice * P.dw) * z;
            epcsi2d_conj(ndxSlice, P, cmplxSlice);

            for (t = 0;t < P.dt;++t)
            {
                for (y = 0;y < P.dy;++y)
                {
                    for (x = 0;x < P.dx; ++x)
                    {
                        ndx1 = t * (2 * P.dx * P.dy) + y * 2 * P.dx;
                        ndx2 = w * (P.dx * P.dy * P.dz * P.dt) + t * (P.dx * P.dy * P.dz) + z * (P.dx * P.dy) + y * P.dx + x;
                        phase5d[ndx2] = atan(cmplxSlice[ndx1+2*x] / cmplxSlice[ndx1+2*x+1]);
                        mag5d[ndx2] = sqrt(cmplxSlice[ndx1+2*x] * cmplxSlice[ndx1+2*x] + cmplxSlice[ndx1+2*x+1] * cmplxSlice[ndx1+2*x+1]);
                    }
                }
            }
        }
    }
    free(cmplxSlice);
}

void flipDim2(double *mat, struct dtiP P)
{
    int i, j, k, t, w;
    int dx, dy, dz, dt, dw, midy, ssize, vsize;
    double tmp[1024];
    long ndx1, ndx2;

    dx = P.dx;
    dy = P.dy;
    dz = P.dz;
    dt = P.dt;
    dw = P.dw;
    vsize = dx*dy*dz;
    ssize = dx*dy;
    midy = dy / 2;

    for (w = 0; w < dw ; ++w)
        for (t = 0; t < dt ; ++t)
            for (k = 0; k < dz ; ++k)
                for (j = 0; j < midy ; ++j)
                {
                    ndx1 = dx * j        + ssize * k + vsize * t + (vsize * dt) * w;
                    ndx2 = dx * (dy - j - 1) + ssize * k + vsize * t + (vsize * dt) * w;

                    for (i = 0; i < dx; ++i)
                    {
                        tmp[i] = mat[ndx1+i];
                    }

                    for (i = 0; i < dx; ++i)
                    {
                        mat[ndx1+i] = mat[ndx2+i];
                    }

                    for (i = 0; i < dx; ++i)
                    {
                        mat[ndx2+i] = tmp[i];
                    }
                }

}

double* GE_Calibration(const char* filename, struct dtiP *Pout)
{
    int dim[] = {64, 64, 1, 64, 7};
    double dTE = 0.000704;
    struct dtiP P;
    short *vec;
    double *phase5d, *mag5d, *mask;
    double *GradAll;
    int i;
    const char outfname1[] = ".\\phase5d.dat";
    const char outfname2[] = ".\\mag5d.dat";
    const char outfname3[] = ".\\grad.txt";
    const char outfname4[] = ".\\log.txt";

    P.dx = dim[0];
    P.dy = dim[1];
    P.dz = dim[2];
    P.dt = dim[3];
    P.dw = dim[4];
    P.dTE = dTE;

    vec = readGEPfile_int16B(filename, P);
    if (vec == NULL)
    {
        printf("Error: Data read in failed. \n");
        return NULL;
    }

    phase5d = malloc(P.dx*P.dy*P.dt*P.dw*P.dz * sizeof(double));
    if (phase5d == NULL)
    {
        printf("Error: Data read in failed. \n");
        free(vec);
        return NULL;
    }
    mag5d   = malloc(P.dx*P.dy*P.dt*P.dw*P.dz * sizeof(double));
    if (mag5d == NULL)
    {
        printf("Error: Data read in failed. \n");
        free(vec);
        free(phase5d);
        return NULL;
    }

    phaserecon(vec, P, phase5d, mag5d);
    free(vec);

    flipDim2(mag5d, P);
    flipDim2(phase5d, P);

    writeFile_double(outfname1 , phase5d, 5 , dim);
    writeFile_double(outfname2 , mag5d, 5 , dim);

    mask = magMask2(mag5d, P.dx, P.dy, P.dz, P.dt);
    free(mag5d);
    if (mask == NULL)
    {
        printf("Error: Data read in failed. \n");
        free(phase5d);
        return NULL;
    }

    GradAll = dti_calibration2(phase5d, P, mask);

    for (i = 0; i < (4*P.dz*(P.dw - 1)); ++i) GradAll[i] *= -1.0;

    free(phase5d);
    free(mask);

    Pout->dx = dim[0];
    Pout->dy = dim[1];
    Pout->dz = dim[2];
    Pout->dt = dim[3];
    Pout->dw = dim[4];
    Pout->dTE = dTE;

    dim[0] = 4;
    dim[1] = dim[2]*(dim[4] - 1);
    writeFile_double2text(outfname3, GradAll, 2, dim, 0);

    return GradAll;
}

int bmuav(double a[] , int m, int n, double u[], double v[], double eps, int ka)
{
    int i, j, k, l, it, ll, kk, ix, iy, mm, nn, iz, m1, ks;
    double d, dd, t, sm, sm1, em1, sk, ek, b, c, shh, fg[2], cs[2];
    double *s, *e, *w;
    void ppp();
    void sss();
    s = malloc(ka * sizeof(double));
    e = malloc(ka * sizeof(double));
    w = malloc(ka * sizeof(double));
    it = 60;
    k = n;
    if (m - 1 < n) k = m - 1;
    l = m;
    if (n - 2 < m) l = n - 2;
    if (l < 0) l = 0;
    ll = k;
    if (l > k) ll = l;
    if (ll >= 1)
    {
        for (kk = 1; kk <= ll; kk++)
        {
            if (kk <= k)
            {
                d = 0.0;
                for (i = kk; i <= m; i++)
                {
                    ix = (i - 1) * n + kk - 1;
                    d = d + a[ix] * a[ix];
                }
                s[kk-1] = sqrt(d);
                if (s[kk-1] != 0.0)
                {
                    ix = (kk - 1) * n + kk - 1;
                    if (a[ix] != 0.0)
                    {
                        s[kk-1] = fabs(s[kk-1]);
                        if (a[ix] < 0.0) s[kk-1] = -s[kk-1];
                    }
                    for (i = kk; i <= m; i++)
                    {
                        iy = (i - 1) * n + kk - 1;
                        a[iy] = a[iy] / s[kk-1];
                    }
                    a[ix] = 1.0 + a[ix];
                }
                s[kk-1] = -s[kk-1];
            }
            if (n >= kk + 1)
            {
                for (j = kk + 1; j <= n; j++)
                {
                    if ((kk <= k) && (s[kk-1] != 0.0))
                    {
                        d = 0.0;
                        for (i = kk; i <= m; i++)
                        {
                            ix = (i - 1) * n + kk - 1;
                            iy = (i - 1) * n + j - 1;
                            d = d + a[ix] * a[iy];
                        }
                        d = -d / a[(kk-1)*n+kk-1];
                        for (i = kk; i <= m; i++)
                        {
                            ix = (i - 1) * n + j - 1;
                            iy = (i - 1) * n + kk - 1;
                            a[ix] = a[ix] + d * a[iy];
                        }
                    }
                    e[j-1] = a[(kk-1)*n+j-1];
                }
            }
            if (kk <= k)
            {
                for (i = kk; i <= m; i++)
                {
                    ix = (i - 1) * m + kk - 1;
                    iy = (i - 1) * n + kk - 1;
                    u[ix] = a[iy];
                }
            }
            if (kk <= l)
            {
                d = 0.0;
                for (i = kk + 1; i <= n; i++)
                    d = d + e[i-1] * e[i-1];
                e[kk-1] = sqrt(d);
                if (e[kk-1] != 0.0)
                {
                    if (e[kk] != 0.0)
                    {
                        e[kk-1] = fabs(e[kk-1]);
                        if (e[kk] < 0.0) e[kk-1] = -e[kk-1];
                    }
                    for (i = kk + 1; i <= n; i++)
                        e[i-1] = e[i-1] / e[kk-1];
                    e[kk] = 1.0 + e[kk];
                }
                e[kk-1] = -e[kk-1];
                if ((kk + 1 <= m) && (e[kk-1] != 0.0))
                {
                    for (i = kk + 1; i <= m; i++) w[i-1] = 0.0;
                    for (j = kk + 1; j <= n; j++)
                        for (i = kk + 1; i <= m; i++)
                            w[i-1] = w[i-1] + e[j-1] * a[(i-1)*n+j-1];
                    for (j = kk + 1; j <= n; j++)
                        for (i = kk + 1; i <= m; i++)
                        {
                            ix = (i - 1) * n + j - 1;
                            a[ix] = a[ix] - w[i-1] * e[j-1] / e[kk];
                        }
                }
                for (i = kk + 1; i <= n; i++)
                    v[(i-1)*n+kk-1] = e[i-1];
            }
        }
    }
    mm = n;
    if (m + 1 < n) mm = m + 1;
    if (k < n) s[k] = a[k*n+k];
    if (m < mm) s[mm-1] = 0.0;
    if (l + 1 < mm) e[l] = a[l*n+mm-1];
    e[mm-1] = 0.0;
    nn = m;
    if (m > n) nn = n;
    if (nn >= k + 1)
    {
        for (j = k + 1; j <= nn; j++)
        {
            for (i = 1; i <= m; i++)
                u[(i-1)*m+j-1] = 0.0;
            u[(j-1)*m+j-1] = 1.0;
        }
    }
    if (k >= 1)
    {
        for (ll = 1; ll <= k; ll++)
        {
            kk = k - ll + 1;
            iz = (kk - 1) * m + kk - 1;
            if (s[kk-1] != 0.0)
            {
                if (nn >= kk + 1)
                    for (j = kk + 1; j <= nn; j++)
                    {
                        d = 0.0;
                        for (i = kk; i <= m; i++)
                        {
                            ix = (i - 1) * m + kk - 1;
                            iy = (i - 1) * m + j - 1;
                            d = d + u[ix] * u[iy] / u[iz];
                        }
                        d = -d;
                        for (i = kk; i <= m; i++)
                        {
                            ix = (i - 1) * m + j - 1;
                            iy = (i - 1) * m + kk - 1;
                            u[ix] = u[ix] + d * u[iy];
                        }
                    }
                for (i = kk; i <= m; i++)
                {
                    ix = (i - 1) * m + kk - 1;
                    u[ix] = -u[ix];
                }
                u[iz] = 1.0 + u[iz];
                if (kk - 1 >= 1)
                    for (i = 1; i <= kk - 1; i++)
                        u[(i-1)*m+kk-1] = 0.0;
            }
            else
            {
                for (i = 1; i <= m; i++)
                    u[(i-1)*m+kk-1] = 0.0;
                u[(kk-1)*m+kk-1] = 1.0;
            }
        }
    }
    for (ll = 1; ll <= n; ll++)
    {
        kk = n - ll + 1;
        iz = kk * n + kk - 1;
        if ((kk <= l) && (e[kk-1] != 0.0))
        {
            for (j = kk + 1; j <= n; j++)
            {
                d = 0.0;
                for (i = kk + 1; i <= n; i++)
                {
                    ix = (i - 1) * n + kk - 1;
                    iy = (i - 1) * n + j - 1;
                    d = d + v[ix] * v[iy] / v[iz];
                }
                d = -d;
                for (i = kk + 1; i <= n; i++)
                {
                    ix = (i - 1) * n + j - 1;
                    iy = (i - 1) * n + kk - 1;
                    v[ix] = v[ix] + d * v[iy];
                }
            }
        }
        for (i = 1; i <= n; i++)
            v[(i-1)*n+kk-1] = 0.0;
        v[iz-n] = 1.0;
    }
    for (i = 1; i <= m; i++)
        for (j = 1; j <= n; j++)
            a[(i-1)*n+j-1] = 0.0;
    m1 = mm;
    it = 60;
    while (1 == 1)
    {
        if (mm == 0)
        {
            ppp(a, e, s, v, m, n);
            free(s);
            free(e);
            free(w);
            return(1);
        }
        if (it == 0)
        {
            ppp(a, e, s, v, m, n);
            free(s);
            free(e);
            free(w);
            return(-1);
        }
        kk = mm - 1;
        while ((kk != 0) && (fabs(e[kk-1]) != 0.0))
        {
            d = fabs(s[kk-1]) + fabs(s[kk]);
            dd = fabs(e[kk-1]);
            if (dd > eps*d) kk = kk - 1;
            else e[kk-1] = 0.0;
        }
        if (kk == mm - 1)
        {
            kk = kk + 1;
            if (s[kk-1] < 0.0)
            {
                s[kk-1] = -s[kk-1];
                for (i = 1; i <= n; i++)
                {
                    ix = (i - 1) * n + kk - 1;
                    v[ix] = -v[ix];
                }
            }
            while ((kk != m1) && (s[kk-1] < s[kk]))
            {
                d = s[kk-1];
                s[kk-1] = s[kk];
                s[kk] = d;
                if (kk < n)
                    for (i = 1; i <= n; i++)
                    {
                        ix = (i - 1) * n + kk - 1;
                        iy = (i - 1) * n + kk;
                        d = v[ix];
                        v[ix] = v[iy];
                        v[iy] = d;
                    }
                if (kk < m)
                    for (i = 1; i <= m; i++)
                    {
                        ix = (i - 1) * m + kk - 1;
                        iy = (i - 1) * m + kk;
                        d = u[ix];
                        u[ix] = u[iy];
                        u[iy] = d;
                    }
                kk = kk + 1;
            }
            it = 60;
            mm = mm - 1;
        }
        else
        {
            ks = mm;
            while ((ks > kk) && (fabs(s[ks-1]) != 0.0))
            {
                d = 0.0;
                if (ks != mm) d = d + fabs(e[ks-1]);
                if (ks != kk + 1) d = d + fabs(e[ks-2]);
                dd = fabs(s[ks-1]);
                if (dd > eps*d) ks = ks - 1;
                else s[ks-1] = 0.0;
            }
            if (ks == kk)
            {
                kk = kk + 1;
                d = fabs(s[mm-1]);
                t = fabs(s[mm-2]);
                if (t > d) d = t;
                t = fabs(e[mm-2]);
                if (t > d) d = t;
                t = fabs(s[kk-1]);
                if (t > d) d = t;
                t = fabs(e[kk-1]);
                if (t > d) d = t;
                sm = s[mm-1] / d;
                sm1 = s[mm-2] / d;
                em1 = e[mm-2] / d;
                sk = s[kk-1] / d;
                ek = e[kk-1] / d;
                b = ((sm1 + sm) * (sm1 - sm) + em1 * em1) / 2.0;
                c = sm * em1;
                c = c * c;
                shh = 0.0;
                if ((b != 0.0) || (c != 0.0))
                {
                    shh = sqrt(b * b + c);
                    if (b < 0.0) shh = -shh;
                    shh = c / (b + shh);
                }
                fg[0] = (sk + sm) * (sk - sm) - shh;
                fg[1] = sk * ek;
                for (i = kk; i <= mm - 1; i++)
                {
                    sss(fg, cs);
                    if (i != kk) e[i-2] = fg[0];
                    fg[0] = cs[0] * s[i-1] + cs[1] * e[i-1];
                    e[i-1] = cs[0] * e[i-1] - cs[1] * s[i-1];
                    fg[1] = cs[1] * s[i];
                    s[i] = cs[0] * s[i];
                    if ((cs[0] != 1.0) || (cs[1] != 0.0))
                        for (j = 1; j <= n; j++)
                        {
                            ix = (j - 1) * n + i - 1;
                            iy = (j - 1) * n + i;
                            d = cs[0] * v[ix] + cs[1] * v[iy];
                            v[iy] = -cs[1] * v[ix] + cs[0] * v[iy];
                            v[ix] = d;
                        }
                    sss(fg, cs);
                    s[i-1] = fg[0];
                    fg[0] = cs[0] * e[i-1] + cs[1] * s[i];
                    s[i] = -cs[1] * e[i-1] + cs[0] * s[i];
                    fg[1] = cs[1] * e[i];
                    e[i] = cs[0] * e[i];
                    if (i < m)
                        if ((cs[0] != 1.0) || (cs[1] != 0.0))
                            for (j = 1; j <= m; j++)
                            {
                                ix = (j - 1) * m + i - 1;
                                iy = (j - 1) * m + i;
                                d = cs[0] * u[ix] + cs[1] * u[iy];
                                u[iy] = -cs[1] * u[ix] + cs[0] * u[iy];
                                u[ix] = d;
                            }
                }
                e[mm-2] = fg[0];
                it = it - 1;
            }
            else
            {
                if (ks == mm)
                {
                    kk = kk + 1;
                    fg[1] = e[mm-2];
                    e[mm-2] = 0.0;
                    for (ll = kk; ll <= mm - 1; ll++)
                    {
                        i = mm + kk - ll - 1;
                        fg[0] = s[i-1];
                        sss(fg, cs);
                        s[i-1] = fg[0];
                        if (i != kk)
                        {
                            fg[1] = -cs[1] * e[i-2];
                            e[i-2] = cs[0] * e[i-2];
                        }
                        if ((cs[0] != 1.0) || (cs[1] != 0.0))
                            for (j = 1; j <= n; j++)
                            {
                                ix = (j - 1) * n + i - 1;
                                iy = (j - 1) * n + mm - 1;
                                d = cs[0] * v[ix] + cs[1] * v[iy];
                                v[iy] = -cs[1] * v[ix] + cs[0] * v[iy];
                                v[ix] = d;
                            }
                    }
                }
                else
                {
                    kk = ks + 1;
                    fg[1] = e[kk-2];
                    e[kk-2] = 0.0;
                    for (i = kk; i <= mm; i++)
                    {
                        fg[0] = s[i-1];
                        sss(fg, cs);
                        s[i-1] = fg[0];
                        fg[1] = -cs[1] * e[i-1];
                        e[i-1] = cs[0] * e[i-1];
                        if ((cs[0] != 1.0) || (cs[1] != 0.0))
                            for (j = 1; j <= m; j++)
                            {
                                ix = (j - 1) * m + i - 1;
                                iy = (j - 1) * m + kk - 2;
                                d = cs[0] * u[ix] + cs[1] * u[iy];
                                u[iy] = -cs[1] * u[ix] + cs[0] * u[iy];
                                u[ix] = d;
                            }
                    }
                }
            }
        }
    }
    return(1);
}

static void ppp(double a[], double e[], double s[], double v[], int m, int n)
{
    int i, j, p, q;
    double d;
    if (m >= n) i = n;
    else i = m;
    for (j = 1; j <= i - 1; j++)
    {
        a[(j-1)*n+j-1] = s[j-1];
        a[(j-1)*n+j] = e[j-1];
    }
    a[(i-1)*n+i-1] = s[i-1];
    if (m < n) a[(i-1)*n+i] = e[i-1];
    for (i = 1; i <= n - 1; i++)
        for (j = i + 1; j <= n; j++)
        {
            p = (i - 1) * n + j - 1;
            q = (j - 1) * n + i - 1;
            d = v[p];
            v[p] = v[q];
            v[q] = d;
        }
    return;
}

static void sss(double fg[2], double cs[2])
{
    double r, d;
    if ((fabs(fg[0]) + fabs(fg[1])) == 0.0)
    {
        cs[0] = 1.0;
        cs[1] = 0.0;
        d = 0.0;
    }
    else
    {
        d = sqrt(fg[0] * fg[0] + fg[1] * fg[1]);
        if (fabs(fg[0]) > fabs(fg[1]))
        {
            d = fabs(d);
            if (fg[0] < 0.0) d = -d;
        }
        if (fabs(fg[1]) >= fabs(fg[0]))
        {
            d = fabs(d);
            if (fg[1] < 0.0) d = -d;
        }
        cs[0] = fg[0] / d;
        cs[1] = fg[1] / d;
    }
    r = 1.0;
    if (fabs(fg[0]) > fabs(fg[1])) r = cs[1];
    else
        if (cs[0] != 0.0) r = 1.0 / cs[0];
    fg[0] = d;
    fg[1] = r;
    return;
}

int bginv(double a[], int m, int n, double aa[], double eps, double u[], double v[], int ka)
{
    int i, j, k, l, t, p, q, f;
    extern int bmuav();
    i = bmuav(a, m, n, u, v, eps, ka);
    if (i < 0) return(-1);
    j = n;
    if (m < n) j = m;
    j = j - 1;
    k = 0;
    while ((k <= j) && (a[k*n+k] != 0.0)) k = k + 1;
    k = k - 1;
    for (i = 0; i <= n - 1; i++)
        for (j = 0; j <= m - 1; j++)
        {
            t = i * m + j;
            aa[t] = 0.0;
            for (l = 0; l <= k; l++)
            {
                f = l * n + i;
                p = j * m + l;
                q = l * n + l;
                aa[t] = aa[t] + v[f] * u[p] / a[q];
            }
        }
    return(1);
}

int agmiv(double a[], int m, int n, double b[], double x[], double aa[], double eps, double u[], double v[], int ka)
{
    int i, j;
    extern int bginv();
    i = bginv(a, m, n, aa, eps, u, v, ka);
    if (i < 0) return(-1);
    for (i = 0; i <= n - 1; i++)
    {
        x[i] = 0.0;
        for (j = 0; j <= m - 1; j++)
            x[i] = x[i] + aa[i*m+j] * b[j];
    }
    return(1);
}

int lsq(double a[], int m, int n, double b[], double x[])
{
    double *u, *v, *aa;
    int ka, r;
    double eps = 0.000001;

    if (m < n)
    {
        return -1;
    }
    ka = m + 1;

    if ((u = malloc(m * m * sizeof(double))) == NULL)
    {
        return -1;
    }
    if ((v = malloc(n * n * sizeof(double))) == NULL)
    {
        free(u);
        return -1;
    }
    if ((aa = malloc(m * n * sizeof(double))) == NULL)
    {
        free(u);
        free(v);
        return -1;
    }
    r = agmiv(a, m, n, b, x, aa, eps, u, v, ka);
    free(u);
    free(v);
    free(aa);

    return r;
}

double* dti_calibration3d(double* data, struct dtiP P, float difscheme[], double difAmp)
{
    double* MtxDW, *MtxDWcopy;

    int ii, kk, r;
    long vsize;
    double *dBdwi;
    double *colB;
    double X[4];
    double *calib3D;

    vsize = P.dx*P.dy*P.dz;

    if ((MtxDW = malloc(4*(P.dw - 1)*sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        return NULL;
    }
    if ((MtxDWcopy = malloc(4*(P.dw - 1)*sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        free(MtxDW);
        return NULL;
    }
    if ((colB = malloc((P.dw - 1)*sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        free(MtxDWcopy);
        free(MtxDW);
        return NULL;
    }
    if ((calib3D = malloc(vsize*4*sizeof(double))) == NULL)
    {
        printf("Error: Memory allocation failed. \n");
        free(MtxDWcopy);
        free(MtxDW);
        free(colB);
        return NULL;
    }

    for (ii = 0;ii < (P.dw - 1);++ii)
    {
        for (kk = 0;kk < 3;++kk)
        {
            MtxDW[4*ii+kk] = difscheme[3*ii+kk] * difAmp;
        }
        MtxDW[4*ii+3] = 1;
    }

    for (kk = 1; kk < P.dw; ++kk)
    {
        dBdwi = data + kk * vsize;

        for (ii = 0;ii < vsize;++ii)
        {
            dBdwi[ii] = dBdwi[ii] - data[ii];
        }
    }

    if (difAmp < 0.0001) difAmp = 1.0;

    for (kk = 0; kk < vsize; ++kk)
    {
        for (ii = 1;ii < P.dw;++ii)
        {
            colB[ii-1] = data[ii*vsize+kk];
        }

        for (ii = 0;ii < 4*(P.dw - 1);++ii)
        {
            MtxDWcopy[ii] = MtxDW[ii];
        }
        r = lsq(MtxDWcopy , P.dw - 1, 4, colB, X);

        if (r < 0)
        {
            for (ii = 0; ii < 4; ++ii)
            {
                X[ii] = 0;
            }
        }

        for (ii = 0; ii < 4; ++ii)
        {
            calib3D[kk+ii*vsize] = X[ii];
        }

    }

    free(MtxDWcopy);
    free(MtxDW);
    free(colB);
    return calib3D;
}