


#include "fileIO.h"


using namespace std;

/*
//	readHeader:
//		Reads Analyze header files for certain data
//		-inputs - filename - filename of the header file
//		-return	- a pointer to the read header or NULL if it could not be read
*/
dsr* readHeader(const char* filename)
{
    if (filename == NULL)
    {
        cout << "ERROR: NULL filename" << endl;
        return NULL;
    }
    FILE* headerFile;
    if ((headerFile = fopen(filename, "rb")) == NULL)
    {
        cout << "ERROR: Header file, " << filename << ", cannot be opened!" << endl;
        return NULL;
    }
    dsr* header = new dsr;
    if (fread(header, sizeof(dsr), 1, headerFile) != 1)
    {
        cout << "ERROR: Header file, " << filename << ", cannot be read!" << endl;
        return NULL;
    }
    fclose(headerFile);
    if (header->hk.sizeof_hdr != 348)
    {
        //BIG ENDIAN!!!
        swapHeaderBytes(header);
        header->dime.dim[7] = -1;
    }
    return header;
}

/*
//	readHeader:
//		-inputs - filename - filename of the header file
//				- dims - pointer to the array that the dimensions will be returned in
//				- newFilename - pointer to the char array that will hold the filename of the
//								image file
//				- type - the datatype of the data
//		-return	- an integer which holds the number of dimensions of the data set
*/
int readHeader(const char* filename, int* dims, int* type, char* newFilename)
{
    if (filename == NULL)
    {
        cout << "ERROR: NULL filename" << endl;
        return NULL;
    }
    dsr* header;
    short int nDim;

    if ((header = readHeader(filename)) == NULL)
    {
        cout << "ERROR: Unable to retrieve header information from " << filename << "!" << endl;
        return -1;
    }

    nDim = header->dime.dim[0];
    if (header->dime.dim[7] == -1)
        dims[nDim] = -1;
    else
        dims[nDim] = 0;
    for (int i = 0; i < nDim; i++)
        dims[i] = header->dime.dim[i + 1];
    *type = header->dime.datatype;
    newFilename = getImageName(filename);
    delete header;
    return nDim;
}

/*
//	readHeader:
//		Reads Analyze header files for certain data
//		-inputs - filename - filename of the header file
//				- dims - pointer to the array that the dimensions will be returned in
//				- newFilename - pointer to the char array that will hold the filename of the
//								image file
//				- type - the datatype of the data
//				- FOV - the field of view of the data
//		-return	- an integer which holds the number of dimensions of the data set
*/

int readHeader(const char* filename, int* dims, int* type, char* newFilename, double* FOV)
{
    if (filename == NULL)
    {
        cout << "ERROR: NULL filename" << endl;
        return NULL;
    }
    dsr* header;
    short int nDim;
    if ((header = readHeader(filename)) == NULL)
    {
        cout << "ERROR: Unable to retrieve header information " << filename << "!" << endl;
        return -1;
    }
    nDim = header->dime.dim[0];
    if (header->dime.dim[7] == -1)
        dims[nDim] = -1;
    else
        dims[nDim] = 0;
    for (int i = 0; i < nDim; i++)
        dims[i] = header->dime.dim[i + 1];
    int multiplier = 1;
    //TODO: check the units
    FOV[0] = header->dime.pixdim[1] * dims[0] * multiplier;
    FOV[1] = header->dime.pixdim[2] * dims[1] * multiplier;
    FOV[2] = header->dime.pixdim[3] * dims[2] * multiplier;
    *type = header->dime.datatype;
    newFilename = getImageName(filename);
    return nDim;
}

/*
//	getImageName:
//		Changes an Analyze header filename to an image filename
//		-inputs - filename - filename to have its extension changed
//		-return - a character string with the .img extension
*/
char* getImageName(const char* filename)
{
    if (filename == NULL)
    {
        cout << "ERROR: NULL filename" << endl;
        return NULL;
    }
    char* newFilename = new char[200];
    int index = 0;
    while (/*filename[index] != NULL &&*/ index < 195)
    {
        newFilename[index] = filename[index];
        if (filename[index] == '.')
        {
            if (filename[index + 4] != NULL)
            {
                index++;
                continue;
            }
            if ((filename[index + 1] == 'h') && (filename[index + 2] == 'd') && (filename[index + 3] == 'r'))
            {
                newFilename[index + 1] = 'i';
                newFilename[index + 2] = 'm';
                newFilename[index + 3] = 'g';
                newFilename[index + 4] = '\0';
                break;
            }
        }
        index++;
    }
    return newFilename;
}

/*
//	getHeaderName:
//		Changes an Analyze image filename to an header filename
//		-inputs - filename - filename to have its extension changed
//		-return - a character string with the .hdr extension
*/
char* getHeaderName(const char* filename)
{
    if (filename == NULL)
    {
        cout << "ERROR: NULL filename" << endl;
        return NULL;
    }
    char* newFilename = new char[200];
    int index = 0;
    while (/*filename[index] != NULL &&*/ index < 195)
    {
        newFilename[index] = filename[index];
        if (filename[index] == '.')
        {
            if (filename[index + 4] != NULL)
            {
                index++;
                continue;
            }
            if ((filename[index + 1] == 'i') && (filename[index + 2] == 'm') && (filename[index + 3] == 'g'))
            {
                newFilename[index + 1] = 'h';
                newFilename[index + 2] = 'd';
                newFilename[index + 3] = 'r';
                newFilename[index + 4] = '\0';
                break;
            }
        }
        index++;
    }
    return newFilename;
}
/*
//	swapHeaderBytes:
//		Swaps the byte order of a header file
//		-inputs - header - header structure to have bytes swapped
//		-return - nothing...changes header
*/
void swapHeaderBytes(dsr* header)
{
    swapBytes(&(header->hk.sizeof_hdr));
    for (int i = 0; i < 8; i++)
        swapBytes(&(header->dime.dim[i]));
    for (int i = 0; i < 8; i++)
        swapBytes(&(header->dime.pixdim[i]));
    swapBytes(&(header->dime.datatype));
    swapBytes(&(header->dime.vox_offset));
    swapBytes(&(header->dime.funused1));
    swapBytes(&(header->dime.funused2));
    swapBytes(&(header->dime.funused3));
    swapBytes(&(header->dime.cal_max));
    swapBytes(&(header->dime.cal_min));
    swapBytes(&(header->dime.datatype));
    swapBytes(&(header->dime.bitpix));
    swapBytes(&(header->hk.extents));
    swapBytes(&(header->dime.glmax));
    swapBytes(&(header->dime.glmin));
    swapBytes(&(header->dime.cal_max));
    swapBytes(&(header->dime.cal_min));
}

/*
//	readFile
//		reads an image file
//		-inputs - filename - filename to be read
//				- nDim - the number of dimensions
//				- dims - the dimensions
//				- type - the type of data
//		-return - returns a pointer to the data
*/
double* readFile(const char* filename, int nDim, int* dim, int type, bool bigEnd)
{
    if (filename == NULL)
    {
        cout << "ERROR: NULL filename" << endl;
        return NULL;
    }
    double* data;
    int size = 1;
    for (int i = 0; i < nDim; i++)
    {
        size *= dim[i];
    }
    data = new double[size];
    if (data == NULL)
    {
        cout << "ERROR: Out of Memory! Data could not be loaded..." << endl;
        return NULL;
    }
    FILE* theFile;
    if ((theFile = fopen(filename, "rb")) == NULL)
    {
        cout << "File, " << filename << ", cannot be opened" << endl;
        return NULL;
    }
    double readDouble;
    float readFloat;
    short int readShort;
    int readInt;
    unsigned char readChar;
    for (int i = 0; i < size; i++)
    {
        if (type == DT_DOUBLE)
        {
            fread(&readDouble, sizeof(double), 1, theFile);
            if (bigEnd)
            {
                swapBytes(&readDouble);
            }
            data[i] = readDouble;
        }
        else if (type == DT_FLOAT)
        {
            fread(&readFloat, sizeof(float), 1, theFile);
            if (bigEnd)
            {
                swapBytes(&readFloat);
            }
            data[i] = (double)readFloat;
        }
        else if (type == DT_SIGNED_SHORT)
        {
            fread(&readShort, sizeof(short int), 1, theFile);
            if (bigEnd)
            {
                swapBytes(&readShort);
            }
            data[i] = (double)readShort;
        }
        else if (type == DT_SIGNED_INT)
        {
            fread(&readInt, sizeof(int), 1, theFile);
            if (bigEnd)
            {
                swapBytes(&readInt);
            }
            data[i] = (double)readInt;
        }
        else if (type == DT_UNSIGNED_CHAR)
        {
            fread(&readChar, sizeof(char), 1, theFile);
            //if(bigEnd)
            //{
            //	swapBytes(&readChar);
            //}
            data[i] = (double)readChar;
        }
        else
        {
            cout << "Error: datatype not supported." << endl;
            return NULL;
        }

    }

    return data;
}

/*
//	readFile
//		reads an image file
//		-inputs - filename - filename to be read
//				- header - header of the image file
//		-return - returns a pointer to the data
*/
double* readFile(const char* filename, dsr* header)
{
    if (filename == NULL)
    {
        cout << "ERROR: NULL filename" << endl;
        return NULL;
    }
    int nDim = header->dime.dim[0];
    int* dims = new int[8];
    for (int i = 0; i < nDim; i++)
    {
        dims[i] = header->dime.dim[i + 1];
    }
    return readFile(filename, nDim, dims, header->dime.datatype, (header->dime.dim[7] == -1));
}

/*
//	getSize
//		returns the size of data type
//		-inputs - type - the data type
//		-return - the size of the data type
*/
int getSize(int type)
{
    return 8;
}


void swapBytes(short int* num)
{
    *num = (((*num >> 8)) | (*num << 8));
}

void swapBytes(int* num)
{
    *num = (((*num & 0x000000FF) << 24) + ((*num & 0x0000FF00) << 8) +
            ((*num & 0x00FF0000) >> 8) + ((*num & 0xFF000000) >> 24));
}

void swapBytes(float* nm)
{
    int* num = (int*)nm;
    *num = (((*num & 0x000000FF) << 24) + ((*num & 0x0000FF00) << 8) +
            ((*num & 0x00FF0000) >> 8) + ((*num & 0xFF000000) >> 24));
}

void swapBytes(double* num)
{
    int* newPointer = (int*)num;
    int firstHalf = *newPointer;
    *newPointer = (((*(newPointer + 1) & 0x000000FF) << 24) + ((*(newPointer + 1) & 0x0000FF00) << 8) +
                   ((*(newPointer + 1) & 0x00FF0000) >> 8) + ((*(newPointer + 1) & 0xFF000000) >> 24));
    *(newPointer + 1) = (((firstHalf & 0x000000FF) << 24) + ((firstHalf & 0x0000FF00) << 8) +
                         ((firstHalf & 0x00FF0000) >> 8) + ((firstHalf & 0xFF000000) >> 24));
}

/*
//	writeHeader
//		writes a header file for an Analyze image
//		-inputs - filename - name of the header file to be written
//				- nDim - number of dimensions
//				- dim - the dimensions of the data
//				- voxDim - the physical size dimensions
//		-return - integer indicating successful writing or not
*/
int writeHeader(const char* filename, int nDim, int* dim, double* voxDim)
{
    if (filename == NULL)
    {
        cout << "ERROR: NULL filename" << endl;
        return -1;
    }
    dsr* header = initHeader();
    if (header == NULL)
    {
        cout << "ERROR: Out of memory! Header structure could not be allocated" << endl;
        return -1;
    }
    header->dime.dim[0] = nDim;
    header->dime.pixdim[0] = nDim;
    for (int i = 0; i < nDim; i++)
        header->dime.dim[i + 1] = dim[i];
    for (int i = 0; i < nDim; i++)
        header->dime.pixdim[i + 1] = voxDim[i];
    return writeHeader(filename, header);
}

/*
//	writeHeader
//		writes a header file for an Analyze image
//		-inputs - filename - name of the header file to be written
//				- header - the header to be written to file
//		-return - integer indicating successful writing or not
*/
int writeHeader(const char* filename, dsr* header)
{
    if (filename == NULL)
    {
        cout << "ERROR: NULL filename" << endl;
        return -1;
    }
    FILE* headerFile;
    if ((headerFile = fopen(filename, "wb")) == NULL)
    {
        cout << "ERROR: File, " << filename << ", could not be created!" << endl;
        return -1;
    }
    if (fwrite(header, sizeof(dsr), 1, headerFile) != 1)
    {
        cout << "ERROR: File, " << filename << ", could not be written to!" << endl;
        return -1;
    }
    fclose(headerFile);
    return 0;
}

/*
//	writeImageFile
//		writes a Analyze image file
//		-inputs - filename - name of the image file to be written
//				- header - the header of the image file to be written
//				- data - the data to be written
//		-return - integer indicating successful writing or not
*/
int writeImageFile(const char* filename, dsr* header, double* data)
{
    if (filename == NULL)
    {
        cout << "ERROR: NULL filename" << endl;
        return -1;
    }
    FILE* imageFile;
    if ((imageFile = fopen(filename, "wb")) == NULL)
    {
        cout << "ERROR: File, " << filename << ", could not be created!" << endl;
        return -1;
    }
    int dims = 1;
    for (int i = 1; i <= header->dime.dim[0]; i++)
        dims *= header->dime.dim[i];
    if (fwrite(data, sizeof(double) * dims , 1, imageFile) != 1)
    {
        cout << "ERROR: File, " << filename << ", could not be written to!" << endl;
        return -1;
    }
    fclose(imageFile);
    return 0;
}

/*
//	initHeader
//		creates a default header
//		-return - pointer to an initialized, default header
*/
dsr* initHeader()
{
    dsr hdr;
    memset(&hdr, 0, sizeof(struct dsr));
    for (int i = 0; i < 8; i++)
        hdr.dime.pixdim[i] = 0.0;
    hdr.dime.vox_offset = 0.0;
    hdr.dime.funused1 = 0.0;
    hdr.dime.funused2 = 0.0;
    hdr.dime.funused3 = 0.0;
    hdr.dime.cal_max = 0.0;
    hdr.dime.cal_min = 0.0;
    hdr.dime.datatype = DT_DOUBLE;
    hdr.dime.bitpix = 64;
    hdr.dime.dim[0] = 4; /* all Analyze images are taken as 4 dimensional */
    hdr.hk.regular = 'r';
    hdr.hk.sizeof_hdr = sizeof(struct dsr);
    hdr.hk.extents = 16384;
    hdr.dime.dim[1] = 1; /* slice width in pixels */
    hdr.dime.dim[2] = 1; /* slice height in pixels */
    hdr.dime.dim[3] = 1; /* volume depth in slices */
    hdr.dime.dim[4] = 1; /* number of volumes per file */
    hdr.dime.glmax = 5; /* maximum voxel value */
    hdr.dime.glmin = 0; /* minimum voxel value */
    /* Assume zero offset in .img file, byte at which pixel
    data starts in the image file */
    hdr.dime.vox_offset = 0.0;
    /* Planar Orientation; */
    /* Movie flag OFF: 0 = transverse, 1 = coronal, 2 = sagittal
    Movie flag ON: 3 = transverse, 4 = coronal, 5 = sagittal */
    hdr.hist.orient = 0;
    /* up to 3 characters for the voxels units label; i.e. mm., um., cm. */
    strcpy(hdr.dime.vox_units, "mm.");
    /* up to 7 characters for the calibration units label; i.e. HU */
    strcpy(hdr.dime.cal_units, " ");
    /* Calibration maximum and minimum values;
    values of 0.0 for both fields imply that no
    calibration max and min values are used */
    hdr.dime.cal_max = 0.0;
    hdr.dime.cal_min = 0.0;
    return &hdr;
}

/*
//	readCalibFile
//		reads a calibration file
//		-inputs - filename - name of the calibration file to be read
//				- size - size of the calibration file to be read
//		-return - pointer to the calibration data
*/
double* readCalibFile(const char* filename, int size)
{
    if (filename == NULL)
    {
        cout << "ERROR: NULL filename" << endl;
        return NULL;
    }
    ifstream calibFile;
    calibFile.open(filename, ios::in);
    double* data = new double[size];
    for (int i = 0; i < size && !calibFile.eof(); i++)
    {
        char test;
        if ((calibFile >> test) == "#")
        {
            while (test != '\n')
                calibFile >> test;
        }
        else
        {
            calibFile.unget();
        }
        calibFile >> data[i];
    }
    calibFile.close();
    return data;
}

/*
//	writeCalibFile
//		writes a calibration file
//		-inputs - filename - name of the calibration file to be written
//				- size - size of the calibration file to be written
//				- data -pointer to the calibration data
//		-return - integer indicating success or not
*/
int writeCalibFile(const char* filename, int size, double* data)
{
    if (filename == NULL)
    {
        cout << "ERROR: NULL filename" << endl;
        return -1;
    }
    ofstream calibFile;
    calibFile.open(filename, ios::out | ios::trunc);
    for (int i = 0; i < size / 4; i++)
    {
        calibFile.width(15);
        calibFile << data[4 * i];
        calibFile << "\t";
        calibFile.width(15);
        calibFile << data[4 * i + 1];
        calibFile << "\t";
        calibFile.width(15);
        calibFile << data[4 * i + 2];
        calibFile << "\t";
        calibFile.width(15);
        calibFile << data[4 * i + 3];
        calibFile << endl;
    }
    calibFile.close();
    return 0;
}

/*
//	validateHeaderFile
//		checks the file to test if it is a valid Analyze header
//		-inputs - filename - name of the file to be validated
//		-return - a boolean value - true if the file is a valid header file, false otherwise
*/
bool validateHeaderFile(const char* filename)
{
    /*char* ext = strrchr(filename, '.');
    if(strcmp(ext, "";*/
    return true;
}

/*
//	checkExtension
//		checks if the filename has the correct extension
//		-inputs	- filename - the filename to be checked
//				- ext - the string containing the extension(s) to check for
//					-format- *.ext *.nextextension *.next
//		-return - true if the filename has the extension, false otherwise
*/
bool checkExtension(const char* filename, const char* extension)
{
    string fname(filename), ext(extension);
    int begin = 0, end = 0;
    while (end <= ext.length())
    {
        begin = ext.find_first_of("*", begin);
        end = ext.find_first_of(" \0", begin);
        const char* sub = strstr(fname.c_str(), ext.substr(begin + 1, end - begin - 1).c_str());
        if (sub != NULL)
        {
            return true;
        }
        begin = end;
    }
    return false;
}